#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define MAP_SIZE 4096UL
#define MAP_MASK (MAP_SIZE - 1)
#define WORD_SIZE sizeof(unsigned long)
#define WORDS_PER_LINE 4

/* memdump
 *
 * tool for dumping /dev/mem. Implementation entirely based on mmap man page!
 *
 *
 * Usage : memdump offset length
 */
#define handle_error(msg)													\
  do { perror(msg); exit(EXIT_FAILURE); } while (0)

int
main(int argc, char *argv[])
{
  unsigned long *addr;
  int fd;  off_t i, n, offset, map_offset;
  size_t length, map_length;
  ssize_t s;
  
  if (argc != 3) {
	 fprintf(stderr, "%s offset length\n", argv[0]);
	 exit(EXIT_FAILURE);
  }
  
  fd = open("/dev/mem", O_RDWR|O_SYNC);
  if (fd == -1)
	 handle_error("open");
  
  offset = strtoul(argv[1], NULL, 0);  
  length = strtoul(argv[2], NULL, 0);
  
  map_length = length & ~MAP_MASK;
  if(map_length == 0)
	 map_length = MAP_SIZE;

  map_offset = offset & ~MAP_MASK;

  printf("Starting %08x (%08x) length %04x (%04x)\n", offset, map_offset, length, map_length);

  addr = mmap(0, map_length, PROT_READ | PROT_WRITE, MAP_SHARED, fd, map_offset);
  if(addr == (void *) -1)
	 handle_error("mmap");

  addr = addr + (offset - map_offset)/WORD_SIZE; // adjust for mapped offset
  
  for(i = 0; i < length; i+=WORD_SIZE) {
	 if(i % (WORD_SIZE*WORDS_PER_LINE) == 0)
		printf("\n%08x:", i+offset);
	 printf(" %08x", *addr);
	 addr++;
  }
  printf("\n");
  exit(EXIT_SUCCESS);
} /* main */

